﻿using System;
using System.Collections.Generic;
using System.Data.Objects.DataClasses;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Security.Permissions;
using System.Threading.Tasks;
using DeepView.Toolkit.GuidGenerator;
using DeepViewDataModel.Common;
using DeepViewDataModel.DataModel;
using DZConverterLib.DAL;
using pdftron.PDF;

namespace DZConverterLib
{
    public class DZConverter: IDisposable
    {
        //public const string ProjectRoot = @"D:\DeepViewConverter\";
        //public const string DownloadDir_PDF = ProjectRoot + @"\PDFPath";
        public string DownloadDir_PDF;
        private PDFDoc doc = null;
        private EntityCollection<DocPage> pageSet = null;
        public int pageCount = 0;
        public ConverterPDF ConverterItem;
        public event EventHandler<MakeConverterErrorArgs> ConverterMakeError;
        public event EventHandler<EndConverterEventArgs> EndConverter;
      
        private static ImageCodecInfo DefaultImageCodecInfo = GetEncoderInfo("image/png");
        private static EncoderParameters DefaultEncoderParameters = new EncoderParameters(2)
        {
            Param = new EncoderParameter[]{
                new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, 100L),
                new EncoderParameter(System.Drawing.Imaging.Encoder.ColorDepth, 8L)
            }
        };
        public DZConverter()
        {
            pdftron.PDFNet.Initialize("daelim.co.kr(Doftech Corp):CPU:2::W:AMC(20120315):EF6E886F25A414FFB5F8C1F2999CF2DA33DC6C5164315BAF7011B87AF0FA");
            
            pageSet = new EntityCollection<DocPage>();
        }

        private void SendError(int Level,string ErrorCode, string _Exception, bool ThreadStop)
        {
            if (ConverterMakeError != null)
                ConverterMakeError(this, new MakeConverterErrorArgs 
                                            {
                                                ThreadStop = ThreadStop, 
                                                ConverterID = ConverterItem.ID,
                                                _Exception = _Exception,
                                                ErrorCode = ErrorCode,
                                                Level = Level
                                            });
        }

        [FileIOPermission(SecurityAction.LinkDemand)]
        public void GetSingleFile(object obj)
        {
            try
            {
                ConverterItem = obj as ConverterPDF;

                using (DeepViewEntities _deepview = new DeepViewEntities(ConnectStringBuilder.DeepViewConnectionString().ToString()))
                {
                    var _property = _deepview.Properties.Where(pro => pro.Property == ConverterItem.ProjectNo);

                    if (_property.Count() > 0)
                    {
                        DownloadDir_PDF = _property.Where(t => t.Type == PropertiesType.PropertiesType.Const_TileSorceStorage).First().Value;

                        if (!Directory.Exists(DownloadDir_PDF)) // 폴더 없을 시 생성
                        {
                            Directory.CreateDirectory(DownloadDir_PDF);
                        }
                    }
                }

                DownloadFile();
            }
            catch (Exception e)
            {
                SendError(0, "Converter Error ", e.ToString(), true);
            }
        }


        private void SetState(IConverterPDF.ConverterStatus Status)
        {
            using (DeepViewEntities _entity = new DeepViewEntities(ConnectStringBuilder.DeepViewConnectionString().ToString()))
            {
                var _items = _entity.ConverterPDF.Where(converter => converter.ID == this.ConverterItem.ID);
                if (_items.Count() > 0)
                {
                    if (Status == IConverterPDF.ConverterStatus.Create)
                    {
                        _items.First().StartDateTime = DateTime.Now;
                    }

                    _items.First().Status = (int)Status;
                    _entity.SaveChanges();
                }
            }
        }

        private void SetCurrentPage(int CurrentPage, IConverterPDF.ConverterStatus Status)
        {

            using (DeepViewEntities _entity = new DeepViewEntities(ConnectStringBuilder.DeepViewConnectionString().ToString()))
            {
                var _items = _entity.ConverterPDF.Where(converter => converter.ID == this.ConverterItem.ID);
                if (_items.Count() > 0)
                {
                    _items.First().Status = (int)Status;
                    _items.First().CurrentPage = CurrentPage;
                    _entity.SaveChanges();
                }
            }
        }

        System.Net.WebClient _fileFactory = new System.Net.WebClient();

        [FileIOPermission(SecurityAction.LinkDemand)]
        public void DownloadFile()
        {
            string DocUri = ConverterItem.PdfUrl; //PDF 전체 경로
            string FileName = DocUri.Remove(0, DocUri.LastIndexOf("/") + 1);
            string ProjectFolderPath = DownloadDir_PDF + "\\" + ConverterItem.ProjectNo + "_Tile"; //프로젝트 폴더
            string ItemListPath = ProjectFolderPath + "\\" + (System.Convert.ToInt64(ConverterItem.SharepointItemID) / 100).ToString();
            string ItemPath = ItemListPath + "\\" + ConverterItem.SharepointItemID;
            string DownloadFilePath = ItemPath + "\\" + FileName;
          
            #region 폴더 생성
            try
            {
     
                DirectoryInfo _dir = new DirectoryInfo(ItemPath);

                if (_dir.Exists)
                {
                    _dir.Empty(false);
                   
                }
                else
                {
                    Directory.CreateDirectory(ItemPath);
                }

            }
            catch (Exception ex)
            {
                //DeleteDirectoryDirs(ItemPath);
                SendError(0,"폴더삭제 에러",ex.ToString(), false);
            }

            #endregion

            #region 다운로드

            try
            {
              
                _fileFactory.UseDefaultCredentials = true;

                _fileFactory.DownloadFileCompleted +=(snd,evt)=>
                {
                    if (File.Exists(DownloadFilePath) && evt.Error == null)
                    {
                        SetState(IConverterPDF.ConverterStatus.Create);

                        #region 기본 설정
                        doc = new PDFDoc(DownloadFilePath);
                        #endregion
                        System.Diagnostics.Debug.WriteLine("download file completed : " + DocUri);
                        ConvertProcess(DownloadFilePath, ItemPath, FileName);
                    }
                    else
                    {
                        SendError(1, "File Not Found", "File Download Error" + DownloadFilePath, true);
                    }
                };
                System.Diagnostics.Debug.WriteLine("download file : " + DocUri);
                _fileFactory.DownloadFileAsync(new Uri(DocUri), DownloadFilePath);

            #endregion
            }
            catch (Exception ex)
            {
                SendError(1, "File Download Error : Url - " + DocUri + " Download Path : " + DownloadFilePath, ex.ToString(), true);
            }
            finally
            {

            }
        }

        [FileIOPermission(SecurityAction.LinkDemand)]
        public void ConvertProcess(string DownloadFilePath, string ItemPath, string FileName)
        {
          
            ResultSet Result_Check = new ResultSet();
            #region PDF 체크
            try
            {
                Result_Check = Check_PDF();
            }
            catch (Exception ex)
            {
                System.Diagnostics.Debug.WriteLine(ex.ToString());
                Result_Check.Message = ex.ToString();
                Result_Check.result = false;
            }
            
            if (!Result_Check.result)
            {
                SendError(1, "PDF 체크 실패 : 보안설정으로 인한 PDF 체크 실패", Result_Check.Message,false);
            } 
            #endregion

            #region PDF에서 이미지로 변환

            try
            {
                int _pageCount = ConvertImage(ItemPath);
                if (_pageCount == pageSet.Count())
                {
                    SetCurrentPage(_pageCount,IConverterPDF.ConverterStatus.Crop);

                    var _infotoDBResult = SetPageInfoToDB(); // 데이터베이스 입력
                    
                    if (_infotoDBResult != null)
                        SendError(0, "SetPageInfoToDB()", _infotoDBResult, true);
                }
                else
                {
                    SendError(0, "converter Check", "Page Count :" + _pageCount + "   PageSet : " + pageSet.Count(), true);
                    return;
                } 
            }
            catch (Exception ex)
            {
                SendError(0, "converter Check ","if (!ConvertImage(ItemPath, FileName)) " + ex.ToString(), true);
            }

            #endregion

            if (EndConverter != null)
                EndConverter(this, new EndConverterEventArgs { ConverterPDF = ConverterItem});
        }

        public ResultSet Check_PDF()
        {
            ResultSet _result = new ResultSet();
            bool _Security = false;
            bool _isEncrypte = true;
            try
            {

                _Security = doc.InitSecurityHandler();
                _isEncrypte = doc.IsEncrypted();

                //접근권한 체크
                if (!_Security)
                {
                    _result.result = false;
                    _result.Message = "Document authentication Error";
                }
                else
                    if (_isEncrypte)
                    {
                        _result.result = false;
                        _result.Message = "Document is Encrypted";
                    }
                    else
                        _result.result = true;

            }
            catch (Exception ex)
            {
                System.Diagnostics.Debug.WriteLine(ex.ToString());
                _result.Message = ex.ToString();
                _result.result = false;
            }

            return _result;
        }


        #region private Method
       
        private static ImageCodecInfo GetEncoderInfo(String mimeType)
        {
            int j;
            ImageCodecInfo[] encoders;
            encoders = ImageCodecInfo.GetImageEncoders();
            for (j = 0; j < encoders.Length; ++j)
            {
                if (encoders[j].MimeType == mimeType)
                    return encoders[j];
            }
            return null;
        }

        private string ConvertDeepImage(string convertFile)
        {
            string _result = string.Empty;

            ImageFields.ImageFields creator = new ImageFields.ImageFields()
            {
                TileSize = 256,
                TileFormat = ImageFields.ImageFormat.AutoSelect,
                CopyMetadata = false,
                ConversionCopyMetadata = false,
                LowLevelPass = true
            };

            FileInfo _file = new FileInfo(convertFile);

            try
            {
                if (_file.Exists)
                {
                    creator.Create(_file.FullName, _file.FullName.ToString().Replace(".jpg", ".xml"));
                    System.Diagnostics.Debug.WriteLine("!!! DeepImage filename : " + convertFile);
                    _file.Delete();
                }
            }
            catch (Exception ex)
            {
                System.Diagnostics.Debug.WriteLine(ex.ToString());

                if (ex.GetType() == typeof(System.ArgumentException))
                    _result = "DPI Exception.";
                else
                    _result = "Code 102312394";
            }
            return _result;
        }

        object _lock = new object();

        private string PageToDeep(Bitmap Bitmap,string FilePath,int Index)
        {
            string _result = string.Empty;
            string pagePath = FilePath + "\\" + Index;
            string ImageName = pagePath + "\\" + Index + ".jpg";

            try
            {
                lock (_lock)
                {
                    using (MemoryStream _savestream = new MemoryStream())
                    {
                        Bitmap.Save(_savestream, DefaultImageCodecInfo, DefaultEncoderParameters);
                        Bitmap = new Bitmap(_savestream);
                        Directory.CreateDirectory(pagePath);
                        Bitmap.Save(ImageName);
                        Bitmap.Dispose();
                    }

                    _result = ConvertDeepImage(ImageName);
                    SetCurrentPage(Index, IConverterPDF.ConverterStatus.Crop);
                }

            }
            catch (Exception ex)
            {
                _result = ex.ToString();
            }

            System.GC.Collect();
            System.GC.WaitForPendingFinalizers();
            return _result;
        }

        private int ConvertImage(string FilePath)
        {
            int _result = 0;
            
            try
            {
                Bitmap newBmp_;
                pageCount = doc.GetPageCount();

                using (DeepViewEntities _entity = new DeepViewEntities(ConnectStringBuilder.DeepViewConnectionString().ToString()))
                {
                    var _items = _entity.ConverterPDF.Where(converter => converter.ID == this.ConverterItem.ID);
                    if (_items.Count() > 0)
                    {
                        _items.First().TotalPages = pageCount;
                        _entity.SaveChanges();
                    }
                }

                #region 이미지 만들기

                for (PageIterator itr = doc.GetPageIterator(); itr.HasNext(); itr.Next())
                {
                    Page _page = itr.Current();
                    int _pageNo = itr.GetPageNumber();

                    try 
	                {
                         System.Diagnostics.Debug.WriteLine("!!! Put Image : " + _pageNo);

                            using (pdftron.PDF.PDFDraw draw = new pdftron.PDF.PDFDraw())
                            {
                                var widthData = (int)_page.GetPageWidth();
                                int heightData = (int)_page.GetPageHeight();
                                pageSet.Add(new DocPage
                                            {
                                                Id = GuidGenerator.GetUniqueGuid(),
                                                PageWidth = (widthData * 3).ToString(),
                                                PageHeight = (heightData * 3).ToString(),
                                                PageNumber = _pageNo,
                                                PageAngle = 0,
                                            });

                                var rotation = doc.GetPage(_pageNo).GetRotation();
                                draw.SetImageSize((int)(widthData * 3), (int)(heightData * 3), true);
                                //draw.SetImageSize((int)(widthData), (int)(heightData), true);
                                draw.SetAntiAliasing(true);
                                draw.SetImageSmoothing(true);
                                draw.SetThinLineAdjustment(true, true);
                                newBmp_ = draw.GetBitmap(_page);
                                SetCurrentPage(_pageNo,IConverterPDF.ConverterStatus.Create);
                                _Task.Add(Task.Factory.StartNew<string>(() => PageToDeep(newBmp_, FilePath, _pageNo)));
                            
                                if (_pageNo == pageCount || _pageNo % 5 == 0)
                                {
                                    Task.WaitAll(_Task.ToArray());
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            SendError(0, "ConverterImage Error : Page NO -" + _pageNo.ToString(), ex.ToString(), true);
                            return 0;
                        }

                }

                Task.WaitAll(_Task.ToArray());
                _result = _Task.Where(r => r.Result.Count() == 0).Count();
                if (_result != pageCount)
                {
                    var _Ex = from task in _Task
                              let a = task.Result
                              where (task.Result.Count() > 0)
                              select new { ex = string.Join(",", a) };

                    SendError(0, "Converter Error code(21324) : Page Count Error", string.Join(",", _Ex), true);
                    return 0;
                }
                
                _Task.ForEach(f=>f.Dispose());
                _Task.Clear();

                GC.Collect(2);
                GC.Collect(2);
            }
            catch (Exception ex)
            {
                 
                SendError(0, "code 1230saf325 Error :", ex.ToString(), true);
            }
            finally
            {
                doc.Close();
                doc.Dispose();
                pdftron.PDFNet.Terminate();
                doc = null;
            }

            #endregion
            return _result;
        }

        List<Task<string>> _Task = new List<Task<string>>();

        private string ConvertImgToDeep(string convertFileName, int Index)
        {
            string _result = string.Empty;
            FileInfo convertFile = new FileInfo(convertFileName);

            ImageFields.ImageFields creator = new ImageFields.ImageFields
            {
                TileSize = 256,
                TileFormat = ImageFields.ImageFormat.AutoSelect,
                CopyMetadata = false,
                ConversionCopyMetadata = false,
                LowLevelPass = true
            };

            try
            {
                if (convertFile.Exists)
                {
                    creator.Create(convertFile.FullName, convertFile.FullName.ToString().Replace(".jpg", ".xml"));
                    convertFile.Delete();
                    
                }
                else
                {
                    _result = "Code 123412545 Page No :" + Index.ToString() + "     ex : File Not";
                }
            }
            catch (Exception ex)
            {
                System.Diagnostics.Debug.WriteLine(ex.ToString());

                if (ex.GetType() == typeof(System.ArgumentException))
                    _result = "DPI Exception. Page No :" + Index.ToString();
                else
                    _result = "Code 102312394 Page No :" + Index.ToString() + "     ex" + ex.Message;
            }
            System.Diagnostics.Debug.WriteLine("ImageFields End : " + _result);
            return _result;
        }

        private string SetPageInfoToDB()
        {
            try
            {
            
                #region DB에 페이지 정보 입력
           

                using (CI_Entities deepViewEntity = new CI_Entities(ConnectStringBuilder.ProjectCIConnectString(ConverterItem.ProjectNo).ToString()))
                {
                    var _docinfo = deepViewEntity
                                    .DocInfo.Where(info => info.OrginPDFUrl == ConverterItem.PdfUrl);

                    if (_docinfo.Count() > 0)
                    {
                        System.Diagnostics.Stopwatch _watch = new System.Diagnostics.Stopwatch();
                        _watch.Start();
                        var _doc = _docinfo.First();
                        _doc.SharepointItemID = ConverterItem.SharepointItemID;
                        _doc.PageCount = pageCount;

                        pageSet.AsParallel().ForAll(p => p.DocInfo_Id = _doc.Id);

                        var _exceptDocInfo = (from newinfo in pageSet
                                              select new { newinfo.PageNumber, newinfo.DocInfo_Id,newinfo.PageWidth,newinfo.PageHeight})
                                                  .Except
                                                        (from oldinfo in _doc.DocPage
                                                         select new { oldinfo.PageNumber, oldinfo.DocInfo_Id, oldinfo.PageWidth, oldinfo.PageHeight });

                        var _delDocInfo = (from oldinfo in _doc.DocPage
                                           select new { oldinfo.PageNumber, oldinfo.DocInfo_Id, oldinfo.PageWidth, oldinfo.PageHeight })
                                                  .Except
                                                        (from newinfo in pageSet
                                                         select new { newinfo.PageNumber, newinfo.DocInfo_Id, newinfo.PageWidth, newinfo.PageHeight })
                                                         .ToList(); /// 삭제되는 collection은 하나씩 제거 될때마다 수정되기 때문에
                                                                    ///  foreach가 안된다. 그러므로 tolist로 가져와서 작업한다.
                        
                        //using (DeepView_SystemEntities SystemEntity = new DeepView_SystemEntities(ConnectStringBuilder.DeepViewSystemConnectionString().ToString()))
                        //{
                        //    SystemEntity.RemoveBackup_DocInfo.AddObject(new RemoveBackup_DocInfo
                        //    {
                        //        BackupUserID = "ConvertStation",
                        //        BackupTime= DateTime.Now,
                        //        ID = GuidGenerator.GetUniqueGuid(),
                        //        OriginDocInfoID = _doc.Id,
                        //        SharepointItemID = _doc.SharepointItemID,
                        //        OriginPDFUrl= _doc.OrginPDFUrl,
                        //        PageCount= _doc.PageCount,
                        //        ProjectNo = ConverterItem.ProjectNo,
                        //        Status = true
                        //    });


                        //    foreach (var item in _delDocInfo)
                        //    {
                        //        var _delitem = _doc.DocPage.Where(p => p.PageNumber == item.PageNumber).First();

                        //        deepViewEntity.DocPage.DeleteObject(_delitem);
                        //        //deepViewEntity.DocPage.DeleteObject(item);
                        //    }
                        //}
                        foreach (var item in _delDocInfo)
                        {
                            var _delitem = _doc.DocPage.Where(p => p.PageNumber == item.PageNumber).First();

                            deepViewEntity.DocPage.DeleteObject(_delitem);
                            //deepViewEntity.DocPage.DeleteObject(item);
                        }

                        deepViewEntity.SaveChanges();

                        foreach (var exceptitem in _exceptDocInfo)
                        {
                            var _lstPage = deepViewEntity.DocPage.Where(page => page.PageNumber == exceptitem.PageNumber
                                                                                    && page.DocInfo_Id == exceptitem.DocInfo_Id);

                            if (_lstPage.Count() > 0)
                            {
                                var _page = _lstPage.First();
                                _page.PageHeight = exceptitem.PageHeight;
                                _page.PageWidth = exceptitem.PageWidth;
                            }
                            else
                            {
                                _doc.DocPage.Add(pageSet.Where(p=>p.PageNumber == exceptitem.PageNumber).First());
                            }
                        }

                        deepViewEntity.SaveChanges();

                        _watch.Stop();
                        System.Diagnostics.Debug.WriteLine(_watch.ElapsedMilliseconds);
                    }
                    else
                    {

                        Guid guid = GuidGenerator.GetUniqueGuid();
                        DocInfo instace = new DocInfo
                        {
                            Id = guid,
                            OrginPDFUrl = ConverterItem.PdfUrl,
                            SharepointItemID = ConverterItem.SharepointItemID,
                            PageCount = pageCount,
                            DocPage = pageSet
                        };
                        deepViewEntity.DocInfo.AddObject(instace);
                    }

                    deepViewEntity.SaveChanges();
                }
            }
            catch (Exception e)
            {
                return e.ToString();
            }
               
            #endregion

            return null;
        }
    #endregion

        public void Dispose()
        {
            if (doc != null)
            {
                doc.Close();
                doc.Dispose();
                pdftron.PDFNet.Terminate();
                doc = null;
            }

            if (_fileFactory != null)
            {
                _fileFactory.Dispose();
                _fileFactory = null;
            }
               

            GC.Collect();
            GC.WaitForPendingFinalizers();
        }
     
    }

    public enum FinalStatus
    {
        /// <summary>
        /// Final PDF를 만들기 위한 순서에 추가
        /// </summary>
        Insert = 0,

        /// <summary>
        /// 앞의 final이 처리중일때 대기
        /// </summary>
        Wait = 1,

        /// <summary>
        /// 만들고 있는중
        /// 이때 데이터베이스에 Current Page는 만들고 있는 페이지
        /// </summary>
        Create = 2,

        /// <summary>
        /// final pdf 완료시
        /// </summary>
        Seccess = 99,

        /// <summary>
        /// 에러
        /// </summary>
        Error = 88
    }

}

